"""
    ulmo.cuahsi.his_central
    ~~~~~~~~~~~~~~~~~~~~~~~

    `CUAHSI HIS Central`_ web services

    .. _CUAHSI HIS Central: http://his.cuahsi.org/hiscentral.html
"""
from .core import get_services
